import logging
import os
import shutil
from itertools import chain
from pathlib import Path
import re

logger = logging.getLogger("imaspy")

this_dir = os.path.abspath(
    os.path.dirname(__file__)
)  # We need to know where we are for many things


def prepare_data_dictionaries():
    """Build IMAS IDSDef.xml files for each tagged version in the DD repository
    1. Search for saxon or download it
    2. Clone the DD repository (ask for user/pass unless ssh key access is available)
    3. Generate IDSDef.xml and rename to IDSDef_${version}.xml
    4. Zip all these IDSDefs together and include in wheel
    """
    saxon_jar_path = get_saxon()
    repo, origin = get_data_dictionary_repo()
    if repo:
        for tag in repo.tags:
            logger.debug("Building data dictionary version {tag}", tag)
            build_data_dictionary(repo, origin, tag, saxon_jar_path)

        from zipfile import ZipFile, ZIP_DEFLATED
        import glob

        logger.info("Creating zip file of DD versions")

        with ZipFile(
            "data-dictionary/IDSDef.zip",
            mode="w",  # this needs w, since zip can have multiple same entries
            compression=ZIP_DEFLATED,
        ) as dd_zip:
            for filename in glob.glob("data-dictionary/[0-9]*.xml"):
                dd_zip.write(filename)


def get_saxon():
    """Search for saxon*.jar and return the path or download it.
    The DD build only works by having saxon9he.jar in the CLASSPATH (until 3.30.0).
    We will 'cheat' a little bit later by symlinking saxon9he.jar to any version
    of saxon we found.

    Check:
    1. CLASSPATH
    2. `which saxon`
    3. /usr/share/java/*
    4. or download it
    """

    local_saxon_path = os.getcwd() + "/saxon9he.jar"
    if os.path.exists(local_saxon_path):
        logger.debug("Something already at ./saxon9he.jar, not creating anew")
        return local_saxon_path

    saxon_jar_origin = Path(
        find_saxon_classpath()
        or find_saxon_bin()
        or find_saxon_jar()
        or download_saxon()
    )
    if not saxon_jar_origin.name == "saxon9he.jar":
        os.symlink(saxon_jar_origin, local_saxon_path)
        return local_saxon_path
    return str(saxon_jar_origin)


def find_saxon_jar():
    import subprocess

    # This finds multiple versions on my system, but they are symlinked together.
    # take the shortest one.
    jars = [
        path
        for path in Path("/usr/share/java").rglob("*")
        if re.match("saxon(.*).jar", path.name, flags=re.IGNORECASE)
    ]

    if jars:
        saxon_jar_path = min(jars, key=lambda x: len(x.parts))
        return saxon_jar_path


def find_saxon_classpath():
    if "CLASSPATH" in os.environ:
        saxon_jar_path = re.search("[^:]*saxon[^:]*jar", os.environ["CLASSPATH"])
        if saxon_jar_path:
            return saxon_jar_path.group(0)


def find_saxon_bin():
    saxon_bin = shutil.which("saxon")
    if saxon_bin:
        with open(saxon_bin, "r") as file:
            for line in file:
                saxon_jar_path = re.search("[^ ]*saxon[^ ]*jar", line)
                if saxon_jar_path:
                    return saxon_jar_path.group(0)


def download_saxon():
    """Downloads a zipfile containing saxon9he.jar and extract it to the current dir.
    Return the full path to saxon9he.jar"""
    from io import BytesIO
    from zipfile import ZipFile
    from urllib.request import urlopen

    SAXON_PATH = (
        "https://iweb.dl.sourceforge.net/project/saxon/Saxon-HE/9.9/SaxonHE9-9-1-4J.zip"
    )

    resp = urlopen(SAXON_PATH)
    zipfile = ZipFile(BytesIO(resp.read()))
    path = zipfile.extract("saxon9he.jar")
    del zipfile
    return path


def get_data_dictionary_repo():
    try:
        import git  # Import git here, the user might not have it!
    except ModuleNotFoundError:
        logger.warning(
            "Could not find 'git' module, try 'pip install gitpython'. \
            Will not build Data Dictionaries!"
        )
        return False, False

        # We need the actual source code (for now) so grab it from ITER
    dd_repo_url = "ssh://git@git.iter.org/imas/data-dictionary.git"
    dd_repo_path = "data-dictionary"

    logger.info("Trying to pull data dictionary git repo from ITER")

    # Set up a bare repo and fetch the access-layer repository in it
    os.makedirs(dd_repo_path, exist_ok=True)
    try:
        repo = git.Repo(dd_repo_path)
    except git.exc.InvalidGitRepositoryError:
        repo = git.Repo.init(dd_repo_path)
    logger.info("Set up local git repository {!s}".format(repo))

    try:
        origin = repo.remote()
    except ValueError:
        origin = repo.create_remote("origin", url=dd_repo_url)
    logger.info("Set up remote '{!s}' linking to '{!s}'".format(origin, origin.url))

    try:
        origin.fetch("--tags")
    except git.exc.GitCommandError:
        logger.warning("Could not fetch tags from %s", list(origin.urls))
    logger.info("Remote tags fetched")
    return repo, origin


def build_data_dictionary(repo, origin, tag, saxon_jar_path, rebuild=False):
    """Build a single version of the data dictionary given by the tag argument
    if the IDS does not already exist.

    In the data-dictionary repository sometimes IDSDef.xml is stored
    directly, in which case we do not call make.
    """
    if (
        os.path.exists("data-dictionary/{version}.xml".format(version=tag))
        and not rebuild
    ):
        return

    import subprocess

    repo.git.checkout(tag)
    # this could cause issues if someone else has added or left IDSDef.xml
    # in this directory. However, we go through the tags in order
    # so 1.0.0 comes first, where git checks out IDSDef.xml
    if not os.path.exists("data-dictionary/IDSDef.xml"):
        try:
            subprocess.check_output(
                "make IDSDef.xml 2>/dev/null",
                cwd=os.getcwd() + "/data-dictionary",
                shell=True,
                env={"CLASSPATH": saxon_jar_path, "PATH": os.environ["PATH"]},
            )
        except subprocess.CalledProcessError:
            logger.warning("Error making DD version {version}", version=tag)
    # copy and delete original instead of move (to follow symlink)
    try:
        shutil.copy(
            "data-dictionary/IDSDef.xml",
            "data-dictionary/{version}.xml".format(version=tag),
            follow_symlinks=True,
        )
    except shutil.SameFileError:
        pass
    os.remove("data-dictionary/IDSDef.xml")


def prepare_ual_sources(ual_symver, ual_commit, force=False):
    """Use gitpython to grab AL sources from ITER repository


    Args:
      - ual_symver: The 'symver style' version to be pulled. e.g. 4.8.2
      - ual_commit: The exact commit to be pulled. Should be a hash or equal to
                    the symver
    """
    try:
        import git  # Import git here, the user might not have it!
    except ModuleNotFoundError:
        logger.warning(
            "Could not find 'git' module, try 'pip install gitpython'. Will not build AL!"
        )
        return False

    # This will probably _always_ depend on the UAL version.
    # However, opposed to the original Python HLI, it does not
    # depend on the IMAS DD version, as that is build dynamically in runtime

    # Now we know which UAL target the user wants to install
    # We need the actual source code (for now) so grab it from ITER
    ual_repo_path = "src/ual"
    ual_repo_url = "ssh://git@git.iter.org/imas/access-layer.git"

    logger.info(
        "Trying to pull commit {!r} symver {!r} "
        "from the {!r} repo at {!r}".format(
            ual_commit, ual_symver, ual_repo_path, ual_repo_url
        )
    )

    # Set up a bare repo and fetch the access-layer repository in it
    os.makedirs(ual_repo_path, exist_ok=True)
    try:
        repo = git.Repo(ual_repo_path)
    except git.exc.InvalidGitRepositoryError:
        repo = git.Repo.init(ual_repo_path)
    logger.info("Set up local git repository {!s}".format(repo))

    try:
        origin = repo.remote()
    except ValueError:
        origin = repo.create_remote("origin", url=ual_repo_url)
    logger.info("Set up remote '{!s}' linking to '{!s}'".format(origin, origin.url))

    origin.fetch("--tags")
    logger.info("Remote tags fetched")

    # First check if we have the commit already
    head = None
    for local_head in repo.heads:
        if local_head.name == ual_commit:
            head = local_head
            logger.info(
                "Found existing head in local repo with commit {!s}".format(head.commit)
            )
            break

    if head is None:
        logger.info("Commit '{!s}' not found locally, trying remote".format(ual_commit))
        # If we do not have the commit, fetch master
        # refspec='remotes/origin/' + ual_commit + ':' + ual_commit
        # refspec = ual_commit + ':' + ual_commit
        refspec = "master"
        logger.info("Fetching refspec {!s}".format(refspec))

        fetch_results = origin.fetch(refspec=refspec)
        if len(fetch_results) == 1:
            head = repo.create_head("HEAD", ual_commit, force=True)
        else:
            raise Exception(
                "Could not create head HEAD from commit '{!s}'".format(ual_commit)
            )

    # Check out remote files locally
    head.checkout()
    described_version = repo.git.describe()
    if described_version != ual_symver:
        raise Exception(
            "Local described version {!r} does"
            " not match requested symver {!r}".format(described_version, ual_symver)
        )

    # We should now have the Python HLI files, check
    hli_src = os.path.join(this_dir, ual_repo_path, "pythoninterface/src/imas")
    if not os.path.isdir(hli_src):
        raise Exception(
            "Python interface src dir does not exist. Should have failed earlier"
        )

    # For the build, we need these
    ual_cython_filelist = [
        "_ual_lowlevel.pyx",
        "ual_defs.pxd",
        "ual_lowlevel_interface.pxd",
    ]
    # We need these in runtime, so check them here
    filelist = ["imasdef.py", "hli_utils.py", "hli_exception.py"]

    # Copy these files into the imaspy directory
    # TODO: This is a bit hacky, do this nicer
    imaspy_libs_dir = os.path.join(this_dir, "imas")
    os.makedirs(imaspy_libs_dir, exist_ok=True)
    # if len(os.listdir(imaspy_libs_dir)) != 0:
    # raise Exception('imaspy libs dir not empty, refusing to overwrite')
    # Make _libs dir act as a python module
    open(os.path.join(imaspy_libs_dir, "__init__.py"), "w").close()

    for file in chain(ual_cython_filelist, filelist):
        path = os.path.join(hli_src, file)
        if not os.path.isfile(path):
            raise Exception(
                "Could not find {!s}, should have failed earlier".format(path)
            )
        else:
            target_path = os.path.join(imaspy_libs_dir, file)
            # Patch some imports, they are different from regular Python HLI and imaspy
            # From PEP-8, absolute imports are preferred https://www.python.org/dev/peps/pep-0008/#id23
            if file == "_ual_lowlevel.pyx":
                with open(path, "r") as old, open(target_path, "w") as new:
                    for line in old:
                        if line == "cimport ual_lowlevel_interface as ual\n":
                            new.write(
                                "cimport imas.ual_lowlevel_interface as ual\n"
                            )
                        elif line == "from imasdef import *\n":
                            new.write("from imas.imasdef import *\n")
                        elif line == "from hli_exception import ALException \n":
                            new.write(
                                "from imas.hli_exception import ALException\n"
                            )
                        else:
                            new.write(line)
            else:
                shutil.copyfile(path, os.path.join(imaspy_libs_dir, file))

    return True


### IMAS-style package names (Not use right now)
def get_ext_filename_without_platform_suffix(filename):
    """ Remove specific system filename extension """
    name, ext = os.path.splitext(filename)
    ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")

    if ext_suffix == ext:
        return filename

    ext_suffix = ext_suffix.replace(ext, "")
    idx = name.find(ext_suffix)

    if idx == -1:
        return filename
    else:
        return name[:idx] + ext


def no_cythonize(extensions, **_ignore):
    """Do not use cython, to generate .c files for this extention
    From https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html
    """
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions



"""
Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import sys
import datetime
from urllib.parse import (
    urljoin,
)

# Sphinx extention to format xarray/pandas summaries
import sphinx_autosummary_accessors

from jinja2.defaults import DEFAULT_FILTERS

from pkg_resources import parse_version as V

import imaspy


print("python exec:", sys.executable)
print("sys.path:", sys.path)

# -- Project information -----------------------------------------------------
# The documented project’s name
project = src_project = PROJECT = "IMASPy"

# PROJECT_ID=10189354
PACKAGE = "imaspy"
PACKAGE_HOST = "gitlab"
src_group = GROUP = "KLIMEX"

# A copyright statement in the style '2008, Author Name'.
copyright = f"2016-{datetime.datetime.now().year}, Karel van de Plassche (DIFFER)"
# The author name(s) of the document
author = "Karel van de Plassche (DIFFER)"
src_host = "gitlab.com"

# Parse urls here for convenience, to be re-used
# gitlab imaspy folder
repository_url = f"https://{src_host}/{src_group}/{src_project}/"
blob_url = urljoin(repository_url, "-/blob/master/")
issue_url = urljoin(repository_url, "-/issues/")
mr_url = urljoin(repository_url, "-/merge_requests/")

# JINTRAC docs
jintrac_sphinx = "https://users.euro-fusion.org/pages/data-cmg/wiki/"

# netCDF4 docs
netcdf4_docs = "https://unidata.github.io/netcdf4-python/netCDF4/index.html"

# ITER docs
iter_projects = "https://git.iter.org/projects/"
imas_site = urljoin(iter_projects, "IMAS/")
imex_site = urljoin(iter_projects, "IMEX/")
al_cython_url = urljoin(imas_site, "repos/al-cython/")
al_python_hli_url = urljoin(imas_site, "repos/al-python/")
al_python_lib_url = urljoin(imas_site, "repos/al-python-lib/")
issue_url = jira_url = "https://jira.iter.org/browse/"

# Configuration of sphinx.ext.extlinks
# See https://www.sphinx-doc.org/en/master/usage/extensions/extlinks.html
# unique name: (base URL, label prefix)
extlinks = {
    "src": (blob_url + "%s", f"{src_group}/{src_project}/"),
    "issue": (issue_url + "%s", None),
    "merge": (mr_url + "%s", "!"),
    "jintrac": (jintrac_sphinx + "%s", "jintrac pages "),
    "netcdf4": (netcdf4_docs + "%s", "netcdf4 "),
    "al_cython": (al_cython_url + "%s", "al_cython "),
    "al_hli": (al_python_hli_url + "%s", "al_hli"),
    "al_lib": (al_python_lib_url + "%s", "al_lib"),
    "pypa": ("https://packaging.python.org/%s", None),
}

full_version = V(imaspy.__version__)

# version: The major project version, used as the replacement for |version|.
#   For example, for the Python documentation, this may be something like 2.6.
version = full_version.base_version

# release: The full project version, used as the replacement for |release| and
#   e.g. in the HTML templates. For example, for the Python documentation, this
#   may be something like 2.6.0rc1
release = str(full_version)


# -- General configuration ---------------------------------------------------
# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",  # To auto-generate docs from Python docstrings
    "sphinx.ext.todo",  # nature theme
    "sphinx.ext.githubpages",  # nature theme
    "sphinx.ext.napoleon",  # Support for NumPy and Google style docstrings
    "sphinx.ext.intersphinx",  # Generate links to other documentation files
    # 'sphinx.ext.coverage',  # numpy
    # 'sphinx.ext.doctest',  # numpy
    "sphinx.ext.autosummary",  # For summarizing autodoc-generated files
    "sphinx.ext.extlinks",  # For shortening internal links
    "sphinx.ext.graphviz",  # Draw Graphs in docs
    # 'sphinx.ext.ifconfig',  # numpy
    # 'matplotlib.sphinxext.plot_directive',  # numpy
    # 'IPython.sphinxext.ipythoGn_console_highlighting',  # numpy
    # 'IPython.sphinxext.ipython_directive',  # numpy
    "sphinx.ext.mathjax",  # Render math with mathjax
    "sphinx_rtd_theme",  # Theme
    "recommonmark",  # For markdown support, does not support 'full' CommonMark syntax (yet)!
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates", sphinx_autosummary_accessors.templates_path]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The suffix of source filenames.
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%Y-%m-%d"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {"logo_only": True}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# '<project> v<release> documentation'.
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/dataset-diagram-logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named 'default.css' will overwrite the builtin 'default.css'.
html_static_path = ["_static"]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = today_fmt

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, 'Created using Sphinx' is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, '(C) Copyright ...' is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. '.xhtml').
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "imaspy_doc"


# -- Extension configuration -------------------------------------------------
# from recommonmark.transform import AutoStructify

# app setup hook
# def setup(app):
#    # add a config value for `ifconfig` directives
#    app.add_config_value('python_version_major', str(sys.version_info.major), 'env')
#    app.add_lexer('NumPyC', NumPyLexer)
#    app.add_config_value('recommonmark_config', {
#        #'url_resolver': lambda url: 'http://' + url,#  a function that maps a existing relative position in the document to a http link
#        'enable_auto_toc_tree': True,  # whether enable Auto Toc Tree feature.
#        'auto_toc_tree_section': 'Contents',  # when enabled, Auto Toc Tree will only be enabled on section that matches the title.
#        'enable_math': True,  # whether enable Math Formula
#        'enable_inline_math': True,  # whether enable Inline Math
#        'enable_eval_rst': False,  # whether Embed reStructuredText is enabled.
#        'enable_auto_doc_ref': None,  # Depracated
#    }, True)
#    app.add_transform(AutoStructify)


# Configuration of sphinx.ext.autodoc
# https://www.sphinx-doc.org/en/master/usage/quickstart.html#autodoc
# autodoc_typehints = "none" #xarray

# Configuration of sphinx.ext.autosummary
# https://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html
autosummary_generate = True

# Configuration of recommonmark
# See https://www.sphinx-doc.org/en/master/usage/markdown.html

# Configuration of sphinx.ext.napoleon
# Support for NumPy and Google style docstrings
# See https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html
# napoleon_google_docstring = True
# napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = True
# napoleon_include_special_with_doc = True
# napoleon_use_admonition_for_examples = False
# napoleon_use_admonition_for_notes = False
# napoleon_use_admonition_for_references = False
# napoleon_use_ivar = False
# napoleon_use_param = True
napoleon_use_keyword = True  # Use the "Keyword Args" syntax
# napoleon_use_rtype = True
napoleon_type_aliases = {
    # general terms
    "sequence": ":term:`sequence`",
    "iterable": ":term:`iterable`",
    "callable": ":py:func:`callable`",
    "dict_like": ":term:`dict-like <mapping>`",
    "dict-like": ":term:`dict-like <mapping>`",
    "mapping": ":term:`mapping`",
    "file-like": ":term:`file-like <file-like object>`",
    # special terms
    # 'same type as caller': '*same type as caller*',  # does not work, yet
    # 'same type as values': '*same type as values*',  # does not work, yet
    # stdlib type aliases
    "MutableMapping": "~collections.abc.MutableMapping",
    "sys.stdout": ":obj:`sys.stdout`",
    "timedelta": "~datetime.timedelta",
    "string": ":class:`string <str>`",
    # numpy terms
    "array_like": ":term:`array_like`",
    "array-like": ":term:`array-like <array_like>`",
    "scalar": ":term:`scalar`",
    "array": ":term:`array`",
    "hashable": ":term:`hashable <name>`",
    # matplotlib terms
    "color-like": ":py:func:`color-like <matplotlib.colors.is_color_like>`",
    "matplotlib colormap name": ":doc:matplotlib colormap name <Colormap reference>",
    "matplotlib axes object": ":py:class:`matplotlib axes object <matplotlib.axes.Axes>`",
    "colormap": ":py:class:`colormap <matplotlib.colors.Colormap>`",
    # objects without namespace
    "DataArray": "~xarray.DataArray",
    "Dataset": "~xarray.Dataset",
    "Variable": "~xarray.Variable",
    "ndarray": "~numpy.ndarray",
    "MaskedArray": "~numpy.ma.MaskedArray",
    "dtype": "~numpy.dtype",
    "ComplexWarning": "~numpy.ComplexWarning",
    "Index": "~pandas.Index",
    "MultiIndex": "~pandas.MultiIndex",
    "CategoricalIndex": "~pandas.CategoricalIndex",
    "TimedeltaIndex": "~pandas.TimedeltaIndex",
    "DatetimeIndex": "~pandas.DatetimeIndex",
    "Series": "~pandas.Series",
    "DataFrame": "~pandas.DataFrame",
    "Categorical": "~pandas.Categorical",
    "Path": "~~pathlib.Path",
    # objects with abbreviated namespace (from pandas)
    "pd.Index": "~pandas.Index",
    "pd.NaT": "~pandas.NaT",
}  # TODO: From xarray, improve!

# From xarray, huh?
# napoleon_preprocess_types = True #From xarray, not in docs
# numpydoc_class_members_toctree = True
# numpydoc_show_class_members = False

# Configuration of sphinx.ext.intersphinx
# See https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "numba": ("https://numba.pydata.org/numba-doc/latest", None),
    "matplotlib": ("https://matplotlib.org", None),
    "xarray": ("http://xarray.pydata.org/en/stable/", None),
    "dask": ("https://docs.dask.org/en/latest", None),
    "cython": ("https://cython.readthedocs.io/", None),
    "gitpython": ("https://gitpython.readthedocs.io/en/stable", None),
    # "netcdf4": ("https://unidata.github.io/netcdf4-python/", None),  # netcdf4 does not have an intersphinx mapping
    # "jintrac": ("https://users.euro-fusion.org/pages/data-cmg/wiki/", None) Behind password, so cannot link there
}

# Configuration of sphinx.ext.graphviz
# See https://www.sphinx-doc.org/en/master/usage/extensions/graphviz.html

# Configuration of sphinx.ext.mathjax
# See https://www.sphinx-doc.org/en/master/usage/extensions/math.html#module-sphinx.ext.mathjax

def escape_underscores(string):
    return string.replace("_", r"\_")


def setup(app):
    DEFAULT_FILTERS["escape_underscores"] = escape_underscores



# pylint: disable=wrong-import-position
"""
Packaging settings. Inspired by a minimal setup.py file, the Pandas cython build
and the access-layer setup template.

The reference method of installing is determined by the
[Python Package Authority](https://packaging.python.org/tutorials/installing-packages/)
of which a summary and advanced explanation is on the [IMASPy wiki](https://gitlab.com/imaspy-dev/imaspy/-/wikis/installing)

The installable IMASPy package tries to follow in the following order:
- The style guide for Python code [PEP8](https://www.python.org/dev/peps/pep-0008/)
- The [PyPA guide on packaging projects](https://packaging.python.org/guides/distributing-packages-using-setuptools/#distributing-packages)
- The [PyPA tool recommendations](https://packaging.python.org/guides/tool-recommendations/), specifically:
  * Installing: [pip](https://pip.pypa.io/en/stable/)
  * Environment management: [venv](https://docs.python.org/3/library/venv.html)
  * Dependency management: [pip-tools](https://github.com/jazzband/pip-tools)
  * Packaging source distributions: [setuptools](https://setuptools.readthedocs.io/)
  * Packaging built distributions: [wheels](https://pythonwheels.com/)

On the ITER cluster we handle the environment by using the `IMAS` module load.
So instead, we install packages to the `USER_SITE` there, and do not use
`pip`s `build-isolation`. See [IMAS-584](https://jira.iter.org/browse/IMAS-584)
"""
import argparse
import distutils.sysconfig
import distutils.text_file
import distutils.util
import importlib
import logging
import os
import site
import pkg_resources
import ast

# Allow importing local files, see https://snarky.ca/what-the-heck-is-pyproject-toml/
import sys

# Import other stdlib packages
from distutils.version import LooseVersion as V
from itertools import chain
from pathlib import Path

# Use setuptools to build packages
from setuptools import Extension
from setuptools import __version__ as setuptools_version
from setuptools import find_packages, setup

if sys.version_info < (3, 6):
    sys.exit(
        "Sorry, Python < 3.6 is not supported. Use a different"
        " python e.g. 'module swap python Python/3.6.4-foss-2018a'"
    )


# Check setuptools version before continuing for legacy builds
if V(setuptools_version) < V("42"):
    ee = RuntimeError(
        "Setuptools version outdated. Found"
        f" {V(setuptools_version)} need at least {V('42')}"
    )
    raise ee

# Workaround for https://github.com/pypa/pip/issues/7953
# Cannot install into user site directory with editable source
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]


# Collect env-specific settings
platform = distutils.util.get_platform()  # linux-x86_64
distutils.util.check_environ()

plat_indep_libraries = Path(distutils.sysconfig.get_python_lib())
plat_indep_include = Path(distutils.sysconfig.get_python_inc())

this_file = Path(__file__)
this_dir = this_file.parent.resolve()  # We need to know where we are for many things

package_name = "imaspy"

# Set up 'fancy logging' to display messages to the user
# Import with side-effects, it sets the root logger
loader = importlib.machinery.SourceFileLoader(
    str(this_dir), package_name + "/setup_logging.py"
)
spec = importlib.util.spec_from_loader(loader.name, loader)
setup_logging = importlib.util.module_from_spec(spec)
loader.exec_module(setup_logging)

logger = logging.getLogger("imaspy")

logger.info("pyproject.toml support got added in pip 10. Assuming it is available")

###
# HANDLE USER ENVIRONMENT
###
loader = importlib.machinery.SourceFileLoader(
    str(this_dir), package_name + "/imas_ual_env_parsing.py"
)
spec = importlib.util.spec_from_loader(loader.name, loader)
imas_ual_env_parsing = importlib.util.module_from_spec(spec)
loader.exec_module(imas_ual_env_parsing)

loader = importlib.machinery.SourceFileLoader(str(this_dir), "setup_helpers.py")
spec = importlib.util.spec_from_loader(loader.name, loader)
setup_helpers = importlib.util.module_from_spec(spec)
loader.exec_module(setup_helpers)

# Now that the environment is defined, import the rest of the needed packages

# Prepare DDs if they can be git pulled
setup_helpers.prepare_data_dictionaries()

# Try to grab all necessary environment variables.
# IMAS_PREFIX points to the directory all IMAS components live in
IMAS_PREFIX = os.getenv("IMAS_PREFIX")
if not IMAS_PREFIX or not os.path.isdir(IMAS_PREFIX):
    logger.warning(
        "IMAS_PREFIX is unset or is not a directory. Points to %s. Will not build UAL!",
        IMAS_PREFIX,
    )
    IMAS_PREFIX = "0.0.0"

# Legacy code, we do not need an explicit IMAS_VERSION.
# What is more important, we need the UAL_VERSION we want to link
# to
## Sanity checks, fallback to 0.0.0_0.0.0
# IMAS_VERSION = os.getenv("IMAS_VERSION")
# if not IMAS_VERSION or not len(IMAS_VERSION.split("."))>2:
#    print('IMAS_VERSION is unset or is not formatted as "0.0.0"')
#    IMAS_VERSION = "0.0.0"
# MAJOR = IMAS_VERSION.split(".")[0]
# MINOR = IMAS_VERSION.split(".")[1]

# Grab the UAL_VERSION to build against
UAL_VERSION = os.getenv("UAL_VERSION")
if not UAL_VERSION:
    logger.warning("UAL_VERSION is unset. Will not build UAL!")
    UAL_VERSION = "0.0.0"

(
    ual_symver,
    steps_from_version,
    ual_commit,
) = imas_ual_env_parsing.parse_UAL_version_string(UAL_VERSION)

safe_ual_symver = imas_ual_env_parsing.sanitise_UAL_symver(ual_symver)
ext_module_name = imas_ual_env_parsing.build_UAL_package_name(
    safe_ual_symver, ual_commit
)

# We need source files of the Python HLI UAL library
# to link our build against, the version is grabbed from
# the environment, and they are saved as syubdirectory of
# imaspy
pxd_path = os.path.join(this_dir, "imas")

LANGUAGE = "c"  # Not sure when this is not true
LIBRARIES = "imas"  # We just need the IMAS library
# EXTRALINKARGS = '' # This is machine-specific ignore for now
# EXTRACOMPILEARGS = '' # This is machine-specific ignore for now


###
# Set up Cython build integration
###
### CYTHON COMPILATION ENVIRONMENT

# From https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#multiple-cython-files-in-a-package
REBUILD_LL = False
if REBUILD_LL:
    USE_CYTHON = True
    cython_like_ext = ".pyx" if USE_CYTHON else ".c"

    ###
    # Build extention list
    ###
    extensions = []

    setup_helpers.prepare_ual_sources(ual_symver, ual_commit)
    try:
        import numpy as np
    except ImportError:
        logger.critical("REBUILD_LL is %s, but could not import numpy", REBUILD_LL)
    else:
        extensions = []

    ual_module = Extension(
        name=ext_module_name,
        sources=["imas/_ual_lowlevel.pyx"],  # As these files are copied, easy to find!
        language=LANGUAGE,
        library_dirs=[IMAS_PREFIX + "/lib"],
        libraries=[LIBRARIES],
        # extra_link_args = [ EXTRALINKARGS ],
        # extra_compile_args = [ EXTRACOMPILEARGS ],
        include_dirs=[IMAS_PREFIX + "/include", np.get_include(), pxd_path],
    )
    if "develop" in sys.argv:
        # Make the dir the .so will be put in
        # Somehow not made automatically
        os.makedirs(ext_module_name.split(".")[0], exist_ok=True)
    extensions.append(ual_module)

    ###
    # Set up Cython compilation (or not)
    ###

    if USE_CYTHON:
        try:
            from Cython.Build import cythonize
        except ImportError:
            logger.critical(
                "USE_CYTHON is %s, but could not import Cython",
                USE_CYTHON,
            )
        else:
            extensions = cythonize(extensions)
    else:
        extensions = setup_helpers.no_cythonize(extensions)
else:
    extensions = []


# long_description
# By default, an upload's description will render with reStructuredText. If the description is in an alternate format like Markdown, a package may set the long_description_content_type in setup.py to the alternate format.
with open(os.path.join(this_dir, "README.md"), encoding="utf-8") as file:
    long_description = file.read()

optional_reqs = {}
for req in ["backends_al", "backends_xarray", "core", "examples", "test", "docs"]:
    optional_reqs[req] = distutils.text_file.TextFile(
        this_dir / f"requirements_{req}.txt"
    ).readlines()
install_requires = optional_reqs.pop("core")
# collect all optional dependencies in a "all" target
optional_reqs["all"] = list(chain(*optional_reqs.values()))

if __name__ == "__main__":
    # Legacy setuptools support, e.g. `python setup.py something`
    # See [PEP-0517](https://www.python.org/dev/peps/pep-0517/) and
    # [setuptools docs](https://setuptools.readthedocs.io/en/latest/userguide/quickstart.html#basic-use)
    pyproject: list = distutils.text_file.TextFile("pyproject.toml").readlines()
    requires_line: str = [line for line in pyproject if "requires =" in line][0]
    requires: str = requires_line.split("=", 1)[1]
    setup_requires: list = ast.literal_eval(requires.strip())
    # Rough logic setuptools_scm
    # See https://pypi.org/project/setuptools-scm/
    # or Git projects, the version relies on git describe
    # no distance and clean:
    #     {tag}
    # distance and clean:
    #     {next_version}.dev{distance}+{scm letter}{revision hash}
    # no distance and not clean:
    #     {tag}+dYYYYMMDD
    # distance and not clean:
    #     {next_version}.dev{distance}+{scm letter}{revision hash}.dYYYYMMDD

    # version_scheme: Configures how the local version number is constructed; either an entrypoint name or a callable.
    # version_scheme options:
    # guess-next-dev:
    #   Automatically guesses the next development version (default). Guesses the upcoming release
    #   by incrementing the pre-release segment if present, otherwise by incrementing the micro
    #   segment. Then appends .devN. In case the tag ends with .dev0 the version is not bumped and
    #   custom .devN versions will trigger a error.
    # post-release:
    #   generates post release versions (adds .postN)
    # python-simplified-semver:
    #   Basic semantic versioning. Guesses the upcoming release by incrementing the minor segment
    #   and setting the micro segment to zero if the current branch contains the string 'feature',
    #   otherwise by incrementing the micro version. Then appends .devN. Not compatible with
    #   pre-releases.
    # release-branch-semver:
    #   Semantic versioning for projects with release branches. The same as guess-next-dev
    #   (incrementing the pre-release or micro segment) if on a release branch: a branch whose
    #   name (ignoring namespace) parses as a version that matches the most recent tag up to the
    #   minor segment. Otherwise if on a non-release branch, increments the minor segment and
    #   sets the micro segment to zero, then appends .devN.
    # no-guess-dev:
    #   Does no next version guessing, just adds .post1.devN

    # local_scheme: Configures how the local version number is constructed; either an entrypoint name or a callable.
    # local_scheme options:
    # node-and-date:
    #   adds the node on dev versions and the date on dirty workdir (default)
    # node-and-timestamp:
    #   like node-and-date but with a timestamp of the form {:%Y%m%d%H%M%S} instead
    # dirty-tag:
    #   adds +dirty if the current workdir has changes
    # no-local-version:
    #   omits local version, useful e.g. because pypi does not support it

    # See https://packaging.python.org/specifications/core-metadata/ for allow version strings
    # Example PEP 440 string:
    # Version: 1.0a2

    setup(
        name=package_name,
        description="Pythonic wrappers for the IMAS Access Layer",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://gitlab.com/klimex/imaspy",
        author="Karel van de Plassche",
        author_email="k.l.vandeplassche@differ.nl",
        classifiers=[
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
            "Natural Language :: English",
            "Programming Language :: Python :: 3",
            "Topic :: Utilities",
        ],
        packages=find_packages(),
        # Include files specified by MANIFEST.in
        include_package_data=True,
        # No pyproject.toml for --no-build-installation. Use setup.py instead
        use_scm_version={
            "write_to": package_name + "/version.py",
            'write_to_template': '"{version}"',
            "relative_to": this_file,
            # For tarball installs without metadata (e.g. .git repository)
            "version_scheme": "guess-next-dev",
            "local_scheme": "no-local-version",
            "fallback_version": os.getenv("IMASPY_VERSION", "0.0.0"),
        },
        python_requires=">3.6",
        # Duplicate from pyproject.toml for older setuptools
        setup_requires=setup_requires,
        install_requires=install_requires,
        extras_require=optional_reqs,
        ext_modules=extensions,
    )
